// ---------------------------------------------------------------------------
//
//                     Include Datei fr RFM12 - Module
//
// ---------------------------------------------------------------------------
//  ==========================================================
//  Autor:   Elmar Faber    Version:  0.1    Datum: 20.09.2010
//  ==========================================================
//
//  Quellen:
//
//  nderungen:
//  ===========
//  20.09.2010: Erste Version
//
//

{$IDATA}
Type
  TRxMode = (rx_onlysend, rx_polling, rx_interrupt);
  TDtMode = (dt_ByteMode, dt_CharMode, dt_SnapMode);
  
  TRFM_BytePacket = Array[0..15] Of Byte;
  TRFM_CharPacket = Array[0..15] Of Char;

  TRFM_Status_Register  = (RGIT_FFIT, POR, RGUR_FFOV, WKUP, EXT, LBD, FFEM,
                           ATS_RSSI, DQD, CRL, ATGL, OFFS6,
                           OFFS3, OFFS2, OFFS1, OFFS0);
  TRFM_Status_RegisterE = Array[TRFM_Status_Register] of byte;
  TRFM_Status_RegisterA = Array[0..15] Of Byte;


Const
  RFM_Version : Byte = 1;
  RFM_Buffer_Count : byte = 16;

Var
 {$IFDEF Mega8}
  nINT[@PortB, 0] : Bit;      // nINT
  nSEL[@PortB, 2] : Bit;      // SS     Select slave
  SDI [@PortB, 3] : Bit;      // MOSI
  SDO [@PinB,  4] : Bit;      // MISO
  SCK [@PortB, 5] : Bit;      // SCK
  nIRQ[@PinD,  2] : Bit;      // nIRQ
  dClk[@PinD,  3] : Bit;      // dClk
  sLED[@PortD, 6] : Bit;      // LED
  FSK [@PortD, 7] : Bit;      // FSK
 {$ENDIF}

 {$IFDEF Mega48}
  FSK [@PortD, 7] : Bit;
  nSEL[@PortB, 2] : Bit;      // SS     Select slave
  SDI [@PortB, 3] : Bit;      // MOSI
  SDO [@PinB,  4] : Bit;      // MISO
  SCK [@PortB, 5] : Bit;      // SCK
  nIRQ[@PinD,  2] : Bit;
  dClk[@PinD,  3] : Bit;

  sLED[@PortD, 6] : Bit;
 {$ENDIF}

 {$IFDEF Mega16}
  FSK [@PortB, 3] : Bit;
  nSEL[@PortB, 4] : Bit;      // SS     Select slave
  SDI [@PortB, 5] : Bit;      // MOSI
  SDO [@PinB,  6] : Bit;      // MISO
  SCK [@PortB, 7] : Bit;      // SCK
  nIRQ[@PinD,  2] : Bit;
  dClk[@PinD,  3] : Bit;

  sLED[@PortD, 6] : Bit;
 {$ENDIF}

  RFM_RxMode : TRxMode;
  RFM_DtMode : TDtMode;
  
  RFM_ib     : Byte;
  RFM_iw     : Word;
  RFM_WaitRx : Boolean;
  RFM_B_Full : Boolean;
  RFM_Count  : Byte;

  RFM_Status_Register : TRFM_Status_RegisterA;
  RFM_Status_RegisterE[@RFM_Status_Register] : TRFM_Status_RegisterE; // access by enum

  RFM_Buffer : Array[0..RFM_Buffer_Count-1] Of Byte;
  RFM_FIFO   : Byte;
  
  RFM_BytePacket : TRFM_BytePacket;
  RFM_CharPacket[@RFM_BytePacket] : TRFM_CharPacket;
  
  {$IFDEF defSnapPacket}
  RFM_SnapPacket[@RFM_BytePacket] : TSnapPacket;
  {$EndIf}
  
{--------------------------------------------------------------}
{--------------------------------------------------------------}
procedure Init_SPI_Ports;
{--------------------------------------------------------------}
begin
 {$IFDEF Mega8}
  Incl(DDRB.0);     // nINT
  Incl(DDRB.2);     // nSel
  Incl(DDRB.3);     // SDI
  Excl(DDRB.4);     // SDO
  Incl(DDRB.5);     // SCK
  Excl(DDRD.2);     // nIRQ
  Excl(DDRD.3);     // dCLK
  Incl(DDRD.6);     // LED
  Incl(DDRD.7);     // FSK
 {$ENDIF}

 {$IFDEF Mega48}
  Incl(DDRB.0);     // nINT
  Incl(DDRB.2);     // nSel
  Incl(DDRB.3);     // SDI
  Excl(DDRB.4);     // SDO
  Incl(DDRB.5);     // SCK
  Excl(DDRD.2);     // nIRQ
  Excl(DDRD.3);     // dCLK
  Incl(DDRD.6);     // LED
  Incl(DDRD.7);     // FSK
 {$ENDIF}

 {$IFDEF Mega16}
  Incl(DDRB.0);     // nINT
  Incl(DDRB.3);     // FSK
  Incl(DDRB.4);     // nSel
  Incl(DDRB.5);     // SDI
  Excl(DDRB.6);     // SDO
  Incl(DDRB.7);     // SCK
  Excl(DDRD.2);     // nIRQ
  Excl(DDRD.3);     // dCLK
  Incl(DDRD.6);     // LED
 {$ENDIF}
end Init_SPI_Ports;

{--------------------------------------------------------------}
{          Liest und Schreibt ein Byte gleichzeitig            }
{--------------------------------------------------------------}
Function Spi_Byte_RW(Spi_out : Byte) : Byte;
{--------------------------------------------------------------}
Var
  i, b : Byte;
Begin
  For i := 0 To 7 Do
    If (Spi_out AND $80) <> 0 Then
      Incl(SDI);
    Else
      Excl(SDI);
    EndIf;
    b := b SHL 1;
    if Bit(SDO) = True Then
      b := b OR 1;
    EndIf;
    SCK := 1;
    SCK := 1;
    SCK := 0;
    SCK := 0;
    Spi_out := Spi_out SHL 1;
  EndFor;
  Incl(SDI);
  Return(b);
End Spi_Byte_RW;

// ------------------------------------------------------------
//                  List den FIFO aus
// ------------------------------------------------------------
Function SPI_RD_FIFO : Byte;
// ------------------------------------------------------------
Var
  i,Result : Byte;
Begin
  SCK  := 0;
  SDI  := 0;
  nSEL := 0;
  for i := 0 To 15 Do //skip status bits
    RFM_Status_Register[i] := 0;
    if Bit(SDO) = True Then
      RFM_Status_Register[i] := 1;
    EndIf;
    SCK := 1;
    SCK := 1;
    SCK := 0;
    SCK := 0;
  EndFor;
  Result := 0;
  for i := 0 To 7 Do  //read fifo data byte
    Result := Result SHL 1;
    if Bit(SDO) = True Then
      Result := Result OR 1;
    EndIf;
    SCK := 1;
    SCK := 1;
    SCK := 0;
    SCK := 0;
  EndFor;
  nSEL := 1;
  return(Result);
End SPI_RD_FIFO;

{--------------------------------------------------------------}
{          Liest und Schreibt ein Word gleichzeitig            }
{--------------------------------------------------------------}
Function Spi_Word_RW(Spi_out : Word) : Word;
{--------------------------------------------------------------}
Var
  i : Byte;
  w : Word;
Begin
  nSel := 0;
  For i := 0 To 15 Do
    If (Spi_out AND $8000) <> 0 Then
      Incl(SDI);
    Else
      Excl(SDI);
    EndIf;
    w := w SHL 1;
    if Bit(SDO) = True Then
      w := w OR 1;
    EndIf;
    SCK := 1;
    SCK := 1;
    SCK := 0;
    SCK := 0;
    Spi_out := Spi_out SHL 1;
  EndFor;
  Incl(SDI);
  nSel := 1;
  Return(w);
End Spi_Word_RW;

// ------------------------------------------------------------
//  Liest 18 Bits aus dem Statusregister
// ------------------------------------------------------------
Procedure Read_Status_Register;
// ------------------------------------------------------------
Var
  i : Byte;
Begin
  nSel := 0;
  For i := 0 To 15 Do
    Excl(SDI);
    RFM_Status_Register[i] := 0;
    if Bit(SDO) = True Then
      RFM_Status_Register[i] := 1;
    EndIf;
    SCK := 1;
    SCK := 1;
    SCK := 0;
    SCK := 0;
  EndFor;
  Incl(SDI);
  nSel := 1;
End Read_Status_Register;

// ------------------------------------------------------------
// ------------------------------------------------------------
Procedure SDI_Send(Sdi_byte : Byte);  // wait for TX FIFO IRQ
// ------------------------------------------------------------
Begin
   While Bit(Nirq)=True Do EndWhile;
   // Transmitter Register Write Command:
   // bit15-bit8 = $B8
   // bit7-bit0 (t7-t0) => data byte
   nSel := 0;
   Spi_Byte_RW($B8);
   Spi_Byte_RW(Sdi_Byte);
   nSel := 1;
End Sdi_send;

// ------------------------------------------------------------
// ------------------------------------------------------------
Procedure Rx_On;
// ------------------------------------------------------------
Begin
  RFM_WaitRx := True;
  SPI_Word_RW($82D9);       // power management: enable receiver, dc=on disable clock  output
  SPI_Word_RW($CA81);       // Reset FIFO  ff=off fifo reset
  SPI_Word_RW($CA83);       // Enable FIFO ff=on  fifo fill
End Rx_On;

// ------------------------------------------------------------
// ------------------------------------------------------------
Procedure Rx_Off;
// ------------------------------------------------------------
Begin
  SPI_Word_RW($0000);       // Read status
  SPI_Word_RW($CA81);       // Reset FIFO ff=off fifo reset
  SPI_Word_RW($8201);       // et,es,ex OFF
  Read_Status_Register;     // Read status / clear status bits
End Rx_Off;

// ------------------------------------------------------------
// ------------------------------------------------------------
Procedure Tx_On;
// ------------------------------------------------------------
Begin
  // Incl(FSK);
  SPI_Word_RW($8238);  // et,es,ex ON  // im PingPong: $8239
  SDI_Send($2D);       // Sync frames
  SDI_Send($D4);       //
End Tx_On;

// ------------------------------------------------------------
// ------------------------------------------------------------
Procedure Tx_Off;
// ------------------------------------------------------------
Begin
  SDI_Send($AA);            // Send 2 dummy bytes to empty fifo
  SDI_Send($AA);            //
  SPI_Word_RW($8201);       // et,es,ex OFF
  Read_Status_Register;     // Read status / clear status bits
  // EXCL(FSK);
End Tx_Off;

// ------------------------------------------------------------
// ------------------------------------------------------------
Procedure ClearBuffer;
// ------------------------------------------------------------
Begin
  For RFM_ib := 0 To RFM_Buffer_Count-1 Do
    RFM_Buffer[RFM_ib] := 0;
  EndFor;
  RFM_B_Full := False; // Puffer kann wieder gefllt werden
End ClearBuffer;

// ------------------------------------------------------------
// ------------------------------------------------------------
Procedure Init_FRM12;
// ------------------------------------------------------------
Begin
  Incl(Fsk);
  RFM_B_Full := False;
  RFM_RxMode := Rx_Polling;
  RFM_DtMode := Dt_ByteMode;
  ClearBuffer;
  mDelay(1000);             // RFM12 Zeit zum initialisieren geben
  
//============================================================================
                            // Status Read Command
                            // ===============================================
  // SPI_Word_RW($0000);    // => Status Read / clear status bits
  Read_Status_Register;     // The read command starts with a zero, whereas all
                            // other control commands start with a one. If a
                            // read command is identified, the status bits will
                            // be clocked out on the SDO pin as follows:
//============================================================================
// RGIT  TX register is ready to receive the next byte (Can be cleared by
//       Transmitter Register Write Command )
// FFIT  The number of data bits in the RX FIFO has reached the pre-programmed
//       limit (Can be cleared by any of the FIFO read methods)
// POR   Power-on reset (Cleared after Status Read Command )
// RGUR  TX register under run, register over write (Cleared after Status Read
//       Command )
// FFOV  RX FIFO overflow (Cleared after Status Read Command )
// WKUP  Wake-up timer overflow (Cleared after Status Read Command )
// EXT   Logic level on interrupt pin (pin 16) changed to low (Cleared after
//       Status Read Command )
// LBD   Low battery detect, the power supply voltage is below the
//       pre-programmed limit
// FFEM  FIFO is empty
// ATS   Antenna tuning circuit detected strong enough RF signal
// RSSI  The strength of the incoming signal is above the pre-programmed limit
// DQD   Data quality detector output
// CRL   Clock recovery locked
// ATGL  Toggling in each AFC cycle
// OFFS(6)  MSB of the measured frequency offset (sign of the offset value)
// OFFS(3)-OFFS(0) Offset value to be added to the value of the frequency
//                 control parameter (Four LSB bits)
//============================================================================
                            // Low Battery Detector and Clock Divider Command
                            // ===============================================
  SPI_Word_RW($c0e0);       // $c0e0 = clock 10 MHz, low battery voltage 2.2V,
                            // $c0c0 = clock  5 MHz, low battery voltage 2.2V,
                            // $c060 = clock  2 MHz, low battery voltage 2.2V,
                            // $c000 = clock  1 MHz, low battery voltage 2.2V,
//============================================================================
                            // Configuration Setting Command:
                            // ===============================================
  SPI_Word_RW($80d7);       // b0-b1: 433MHz band b0=1, b1=0
                            // x0-x3: 12pf        x0-x2=1, x3=0
                            // el=On: enable TX register
                            // ef=On: enable RX FIFO buffer
//============================================================================
                            // Frequency Setting Command:
                            // ===============================================
  SPI_Word_RW($a620);       // Frequency: 433,92MHz
//============================================================================
                            // Power Management Command:
                            // ===============================================
  SPI_Word_RW($8201);       // power management:
                            // er = Off => disable >> receiver << chain
                            // es = Off => disable synthesizer
                            // ex = Off => disable crystal oscillator
                            // et = Off => disable PLL, PA and >> TX <<
                            // dc = On  => disable clock output
//============================================================================
                            // Data Rate Command:
                            // ===============================================
  SPI_Word_RW($c611);       // Datarate: $c602 115,2 kbit
                            // Datarate: $c605  57,6 kbit
                            // Datarate: $c608  38,4 kbit
                            // Datarate: $c611  19,2 kbit
                            // Datarate: $c617  14,4 kbit
                            // Datarate: $c623   9,6 kbit
                            // Datarate: $C688   4,8 kbit
                            // Datarate: $c691   2,4 kbit
//============================================================================
                            // Data filter command:
                            // ===============================================
  SPI_Word_RW($c2ac);       // Clock recovery (CR) auto lock control
                            // al = 1 - CR auto lock control
                            //      CR will start in fast mode, then after
                            //      locking it will automatically switch to
                            //      slow mode
                            // Clock recovery lock control
                            // ml = 1 - fast mode
                            //  s = 1 - digital filter
                            // f0-f2: DQD (Data quality detection) threshold
                            //        (Grenzwert/Schwelle)
//============================================================================
                            // Receiver Control Command:
                            // ===============================================
  SPI_Word_RW($94a1);       // receiver setting: 134kHz, -97dbm
                            // VDI (valid data indicator) response time=fast
                            // p20 = On  => VDI output
                            // p20 = Off => INT input (= $90a1)
//============================================================================
                            // FIFO and Reset Mode Command:
                            // ===============================================
  SPI_Word_RW($ca83);       // f3-f0: FIFO IT level. The FIFO generates IT when
                            //        the number of received data bits reaches
                            //        this level
                            // f3=1:  set level to 8 Bit
                            //
                            // al=0:  Set the input of the FIFO fill start
                            //        condition to synchron pattern
                            //        synchron pattern is: 2DD4h
                            // ff=0:  FIFO fill will be enabled after synchron
                            //        pattern reception. The FIFO fill stops
                            //        when this bit is cleared.
                            // dr=1:  Disables the highly sensitive RESET mode.
                            //        If this bit is cleared, a 200 mV glitch in
                            //        the power supply may cause a system reset.
                            // =====
                            // Note: To restart the synchron pattern recognition,
                            //       bit 1 should be cleared and set.
                            // =====
//============================================================================
                            // AFC (Automatic frequency control) Command:
                            // ===============================================
  SPI_Word_RW($c483);       // a1-a0: Automatic operation mode selector:
                            // a1=1,a0=0: Keep the foffset only during receiving
                            //            (VDI=high)
                            // rl1-rl0): Range limit. Limits the value of the
                            //           frequency offset register to the next
                            //           values:
                            // rl1=0,rl0=0: No restriction
                            // st=0:  Strobe edge
                            // fi=0:  Fine mode
                            // oe=1:  Offset register enabled
                            // en=1:  Enables the calculation of the offset
                            //        frequency by the AFC circuit.
//============================================================================
                            // TX Configuration Control Command:
                            // ===============================================
  SPI_Word_RW($9850);       // p2-p0: alles 0 = Output power 0dB
                            // mp, m3 - m0: FSK modulation parameters
                            // mp=0,m3=0,m2=1,m1=0,m0=1:
                            // Delta f (fsk) = 90 kHz
                            //
//============================================================================
                            // Wake-Up Timer Command:
                            // ===============================================
  SPI_Word_RW($e000);       //
//============================================================================
                            // Low Duty-Cycle Command:
                            // ===============================================
  SPI_Word_RW($c800);       //
//============================================================================
  // SPI_Word_RW($0000);    // read status / clear status bits
  Read_Status_Register;
//============================================================================

End;

// ------------------------------------------------------------
// ------------------------------------------------------------
procedure Enable_Int0;
// ------------------------------------------------------------
begin
  asm: sei;            // Setze I-Flag im SREG
  {$IFDEF Mega16}
  MCUCR := MCUCR OR %00000000;  // The low level of INT0 generates an interrupt request.
  GICR  := GICR  OR %01000000;  // Int0 aktiv
  {$ENDIF}
  {$IFDEF Mega8}
  MCUCR := MCUCR OR %00000000;  // The low level of INT0 generates an interrupt request.
  GICR  := GICR  OR %01000000;  // Int0 aktiv
  {$ENDIF}
  {$IFDEF Mega48}
  EICRA := EICRA OR %00000000;  // The low level of INT0 generates an interrupt request.
  EIFR  := EIFR  OR %00000001;  // Int0 aktiv
  {$ENDIF}
  RFM_Count  := 0;
  RFM_B_Full := False;
end;

// ------------------------------------------------------------
// ------------------------------------------------------------
procedure Disable_Int0;
// ------------------------------------------------------------
begin
  {$IFDEF Mega16}
  GICR  := GICR OR %00000000;  // Int0 inaktiv
  {$ENDIF}
  {$IFDEF Mega8}
  GICR  := GICR OR %00000000;  // Int0 inaktiv
  {$ENDIF}
  {$IFDEF Mega48}
  EIFR  := EIFR OR %00000000;  // Int0 inaktiv
  {$ENDIF}
  RFM_Count  := 0;
  RFM_B_Full := False;
end;

// ------------------------------------------------------------
// ------------------------------------------------------------
Procedure Set_RxMode(RxMode : TRxMode; dtMode : TDtMode);
// ------------------------------------------------------------
Begin
  RFM_RxMode := RxMode;
  RFM_DtMode := dtMode;
  Case RFM_RxMode Of
    rx_onlysend  : Disable_Int0;
                   SPI_Word_RW($94a1);
                   |
    rx_polling   : SPI_Word_RW($94a1);
                   Read_Status_Register; // Read status / clear status bits
                   Disable_Int0;
                   |
    rx_interrupt : SPI_Word_RW($94a1);   // SPI_Word_RW($90a1);
                   Read_Status_Register; // Read status / clear status bits
                   Enable_Int0;
                   |
  EndCase;
End Set_RxMode;


// ------------------------------------------------------------
// ------------------------------------------------------------
Interrupt Int0;
// ------------------------------------------------------------
Begin
  IF RFM_B_Full <> True Then
    Incl(sLed);

    For RFM_ib := 0 To RFM_Buffer_Count-1 Do
      While Bit(Nirq)=True Do EndWhile;
      Spi_Byte_RW($B0);
      RFM_Buffer[RFM_ib] := SPI_RD_FIFO;
    EndFor;

    Excl(sLed);
  
    RFM_B_Full := True;

    // Empfang erst einmal abstellen bis Daten verarbeitet wurden
    Read_Status_Register;     // Read status / clear status bits
    SPI_Word_RW($CA81);       // Reset FIFO ff=off fifo reset
    SPI_Word_RW($8201);       // et,es,ex OFF
    Read_Status_Register;     // Read status / clear status bits
  EndIf;
End;

// ------------------------------------------------------------
// ------------------------------------------------------------
Function RFM_Receive_Data : Boolean;
// ------------------------------------------------------------
Var Resu : Boolean;
    CRC8 : Byte;
Begin
  If RFM_RxMode = rx_polling Then
    Excl(sLed);
    For RFM_ib := 0 To RFM_Buffer_Count-1 Do
      While Bit(nIRQ)=True Do
        If RFM_WaitRx = False Then
          writeln(SerOut, 'Break');
          Break;
        EndIf;
      EndWhile;
      If RFM_WaitRx = False Then
        writeln(SerOut, 'Break');
        Break;
      EndIf;
      // Receiver FIFO Read Command: $B000
      // With this command, the controller can read 8 bits from the receiver FIFO.
      // Bit 6 (ef) must be set in Configuration Setting Command.
      Spi_Byte_RW($B0);
      RFM_Buffer[RFM_ib] := SPI_RD_FIFO;
    EndFor;
    Excl(sLed);

    If RFM_WaitRx = False Then
      RFM_B_Full := False;
    Else
      RFM_B_Full := True;
    EndIf;
    
    // Empfang erst einmal abstellen bis Daten verarbeitet wurden

    SPI_Word_RW($0000);       // Read status
    SPI_Word_RW($CA81);       // Reset FIFO ff=off fifo reset
    SPI_Word_RW($8201);       // et,es,ex OFF
    Read_Status_Register;     // Read status / clear status bits
  EndIf;
  
  If RFM_WaitRx = False Then
    Return(False);
  EndIf;

  Resu := True;
  //
  // RFM_B_Full kann auch im Interrupt-Modus True werden!
  // Wenn also wie auch immer Daten angekommen sind, dann
  // Je nach Modus die Daten verarbeiten
  //
  If RFM_B_Full Then
    RFM_BytePacket := RFM_Buffer;
    If RFM_DtMode = dt_SnapMode Then
      {$IFDEF defSnapPacket}
      CRC8 := SnapCRC8(RFM_SnapPacket);
      IF CRC8 <> RFM_SnapPacket.CRC8 Then
        Resu := false;
        RFM_SnapPacket.Data[0] := CRC8;
      Else
        Resu := True;
      EndIf;
      {$EndIf}
    EndIf;
  EndIf;

  Return(Resu);
End RFM_Receive_Data;

// ------------------------------------------------------------
// ------------------------------------------------------------
Procedure SendPacket;
// ------------------------------------------------------------
Var iw : Word;
Begin
  Incl(sLed);
  Case RFM_DtMode Of
    dt_ByteMode : For iw := 0 To SizeOf(RFM_BytePacket)-1 Do
                    Sdi_send(RFM_BytePacket[iw]);
                  EndFor;
                  |
    dt_SnapMode : {$IFDEF defSnapPacket}
                  For iw := 0 To SizeOf(pSnapPacket)-1 Do
                    Sdi_send(pSnapPacket[iw]);
                  EndFor;
                  {$EndIF}
                  |
  EndCase;
  Excl(sLed);
End SendPacket;

