// ---------------------------------------------------------------------------
//
//  Include Datei fr die Verwendung des SNAP-Protokolls
//
//  Autor:   Elmar Faber
//
//  Quellen: Basiert auf Code von Benedikt K. ( in C )
//           Joachim Fossie Br Reiter 04/2007 ( BASCOM )
//           Version: v1.01
//           Author:  Kurzschluss
//
//  nderungen:
//  ===========
//  13.09.2010: Erste Version
//
//
//

// The packet structure is defined in the received packets first two
// bytes (HDB2 and HDB1). The following packet structure is used.
//
// DD=10     - 2 Byte destination address
// SS=10     - 2 Byte source address
// PP=00     - No protocol specific flags
// AA=00     - No Acknowledge is required
// D =0      - No Command Mode
// EEE=011   - 8-bit CRC
// NNNN=1000 - 8 Byte data
//
// Overview of header definition bytes (HDB2 and HDB1)
//
//         HDB2             HDB1
// +-----------------+-----------------+
// | D D S S P P A A | D E E E N N N N |
// +-----------------+-----------------+
// | 1 0 1 0 0 0 0 0 | 0 0 1 1 1 0 0 0 |
// +-----------------+-----------------+

{$Define defSnapProt}

{$IDATA}
Const
  SyncByte : Byte = %01010100;
  HDB2Byte : Byte = %10100000;
  HDB1Byte : Byte = %00111000;
  // Address Bytes:
  // --------------
  // SAB1/DAB1 - Byte = Type of Devive
  // 00 : Unspecific Device
  // 01 : Temperature Device
  // 02 : Data
  // 03 : Clock
  
  ABUnspec : Byte = 00;
  ABTempDv : Byte = 01;
  ABDataDv : Byte = 02;
  ABClockDv: Byte = 03;

  DAB2Byte : Byte = 00;
  SAB2Byte : Byte = 00;

  Snap_S_Address : byte = defSnap_S_Address;
  Snap_T_Address : byte = defSnap_T_Address;

  // SNAP Error Codes
  SN_No_Error  : Byte = 0;
  SN_CRC_Error : Byte = 1;
  SN_ADR_Error : Byte = 2; // Falsche Adresse, nicht fr mich bestimmt...
  SN_UNK_Error : Byte = 3; // Unbekannter Fehler
  SN_SYN_Error : Byte = 4; // Sync-Fehler

{ Type Declarations }
type
  TSnapDataBuffer = Array[0..15] Of Byte;

  TSnapDataArray  = Array[0..7] Of Byte;
  TSnapDataString = String[8];

  TSnapPacket=Record
    SYNC,              // Synchonization Byte
    HDB2,              // Header Definition Byte 2
    HDB1,              // Header Definition Byte 1
    DAB2,              // Destination Address Byte 2
    DAB1,              // Destination Address Byte 1
    SAB2,              // Source Address Byte 2
    SAB1    : Byte;    // Source Address Byte 1
    Data    : TSnapDataArray;
    CRC8    : Byte;
  End;
  
{ Var Declarations }
Var
  Snap_i      : Byte;
  SnapPacket  : TSnapPacket;
  pSnapPacket[@SnapPacket] : TSnapDataBuffer;
  SnapString  : TSnapDataString;


(*
rw := CRCcheck(pSnapPacket, SizeOf(pSnapPacket));

function CrcCheck1(PData : PChar; Cnt: Word): word;
var
  Crc : Word;
  i   : Byte;
begin
  Crc := 0;
  while (Cnt > 0) do
    Crc := Crc xor (Word(PData^) shl 8);
    Inc(PData);
    for i := 1 to 8 do
      if (crc > $7FFF) then
        Crc := (Crc shl 1) xor $1021
      else
        Crc := Crc shl 1;
    endfor;
    Dec(Cnt);
  endWhile;
  Result := Swap(Crc);
end;

*)
  
//----------------------------------------------------------
//----------------------------------------------------------
function SnapCRC8(aPacket : TSnapPacket) : Byte;
//----------------------------------------------------------
var Result,
    j, k, x : Byte;
    m : Integer;
    pPacket[@aPacket] : TSnapDataBuffer;
begin
  Result := 0;
  for m := 0 to SizeOf(pPacket)-1 do
    x := pPacket[m];
    for k := 0 to 7 do
      j := 1 and (x xor Result);
      Result := Result div 2;
      x := x div 2;
      if j <> 0 then
        Result := Result xor $8C;
      EndIf;
    endFor;
  endFor;
  Return(Result);
end SnapCRC8;

(*
// Subroutine for calculating CRC value in variable byte
symbol CRCPOLY		= $1021         ' CRC-CCITT

calc_crc:
        CRC = Byte * 256 ^ CRC
        for temp = 0 to 7
            if Bit15 = 0 then shift_only
            CRC = CRC * 2 ^ CRCPOLY
            goto nxt
shift_only: CRC = CRC * 2
nxt:    next
        return

' -----[ End ]--------------------------------------------------------
*)

//----------------------------------------------------------
//----------------------------------------------------------
Procedure SnapInit;
//----------------------------------------------------------
Begin
  With Snappacket Do
    SYNC := SyncByte;
    HDB2 := HDB2Byte;
    HDB1 := HDB1Byte;
    DAB2 := DAB2Byte;
    DAB1 := ABUnspec;
    SAB2 := SAB2Byte;
    SAB1 := ABUnspec;
    For Snap_i := 0 To 7 Do
      Data[Snap_i] := $FF;
    EndFor;
  EndWith;
  Snappacket.CRC8 := SnapCRC8(Snappacket);
End SnapInit;

//----------------------------------------------------------
//----------------------------------------------------------
Procedure SnapSetSourceAddress(aSAB2, aSAB1 : Byte);
//----------------------------------------------------------
Begin
  With Snappacket Do
    SAB2 := aSAB2;
    SAB1 := aSAB1;
  EndWith;
  Snappacket.CRC8 := SnapCRC8(Snappacket);
End SnapSetSourceAddress;

//----------------------------------------------------------
//----------------------------------------------------------
Procedure SnapSetTargetAddress(aDAB2, aDAB1 : Byte);
//----------------------------------------------------------
Begin
  With Snappacket Do
    DAB2 := aDAB2;
    DAB1 := aDAB1;
  EndWith;
  Snappacket.CRC8 := SnapCRC8(Snappacket);
End SnapSetTargetAddress;

//----------------------------------------------------------
//----------------------------------------------------------
Procedure SnapSetData(SnapData : TSnapDataArray);
//----------------------------------------------------------
Begin
  Snappacket.Data := SnapData;
  Snappacket.CRC8 := SnapCRC8(Snappacket);
End SnapSetData;

//----------------------------------------------------------
//----------------------------------------------------------
Function SnapPacketValidator(aSnapPacket : TSnapPacket) : Byte;
//----------------------------------------------------------
Var
  aCRC : Byte;
Begin
  If aSnapPacket.Sync = SyncByte Then
    // HDB2Byte : Byte = %10100000;
    // HDB1Byte : Byte = %00111000;
    // 2-Byte Adressierung?
    If (Bit(aSnapPacket.Hdb2,7)=True) AND (Bit(aSnapPacket.Hdb2,6)=False) Then
      If aSnapPacket.DAB2 = Snap_S_Address Then // Adresse prfen
        aCrc := SnapCRC8(aSnapPacket);
        // NACK
        Incl(aSnapPacket.HDB2,0);
        Incl(aSnapPacket.HDB2,1);
        If aCrc = aSnapPacket.CRC8 Then          // CRC OK?
          Return(SN_No_Error);
        Else
          Return(SN_CRC_Error);
        EndIf;
      Else
        Return(SN_ADR_Error);
      EndIf;
    EndIf;
  Else
    Return(SN_SYN_Error);
  EndIf;
End SnapPacketValidator;

//----------------------------------------------------------
//----------------------------------------------------------
Function SnapPacketAck(Var aSnapPacket : TSnapPacket) : Boolean;
//----------------------------------------------------------
Var
  aDAB2 : Byte;
Begin
  // Acknowledge senden???
  IF (NOT Bit(aSnapPacket.HDB2,0)) AND Bit(aSnapPacket.HDB2,1) Then
    aDAB2 := aSnapPacket.DAB2;
    aSnapPacket.DAB2 := aSnapPacket.SAB2;
    aSnapPacket.SAB2  := aDAB2;
    Excl(aSnapPacket.HDB2,0);
    Incl(aSnapPacket.HDB2,1);
    Return(True);
  Else
    Return(False);
  EndIf;
End SnapPacketAck;

